################################################################
# Fit a spline [B]w over ages x1...xN to minimize wtd sum of penalties
# (1)  FIT:    small eps = f - GBw
# (2)  SHAPE:  small v = MSBw  (M is projection matrix for HFD comp.)
#
# calculate Kf,Kw, and moment multipliers and write to text files
################################################################

rm(list=ls())
##@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
graphics.off()
windows(record=TRUE)
##@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

library(splines)

#@@@@@@@@@@@
PC        = 3        # number of principal SVD components
Wstar     = 2000      # typical number of women per single-year age (higher value
                      #   means input data less noisy and fitting becomes more impt)

splineDegree = 2      # decided on 2 rather than 3 because HFD fits are SLIGHTLY better

#----------------------------------------
# read prepared f5,f1,SVD,NMF... data
#----------------------------------------
D = dget("fert.data.dput")  
for (k in names(D)) assign(k, D[[k]])

dimnames(BIG.BASIS) = list(12:54, NULL)

alpha = 12
beta  = 55
age   = alpha:(beta-1)
A     = length(age)

#--------------------------------------
# calculate empirical shape residuals over
#   [HFD5 + IDB]
#--------------------------------------

X           = BIG.BASIS[paste(age),1:PC] 
dimnames(X) = list(age,NULL)

PR  = X %*% solve(t(X) %*% X) %*% t(X)  # projection
M   = diag(A) - PR                      # residuals


es   = M %*% f1[paste(age),]
tot = matrix(0,A,A)
for (i in 1:ncol(es)) {
 z = matrix(es[,i], ncol=1)
 tot = tot + 1/ncol(es) * (z %*% t(z))
}

Vsh    = tot

# add constant to diagonal to stabilize empirical estimate before inverting
VshInv = solve( Vsh + 0.1 * median(diag(Vsh))*diag(A))

#--------------------------------------
# set up grid and basis functions
#--------------------------------------

na    = 2
Delta = 1/na

a        = seq(alpha+Delta/2, beta-Delta/2, Delta)
y        = NA*a
names(y) = a
N        = length(y)

knot = seq(alpha, beta, 2 )  

## quad spline basis

B = bs(a, knots=knot, degree= splineDegree)

keep = (splineDegree+1):(ncol(B)-(splineDegree+1))

B    = B[,keep]

nparam = ncol(B)

dimnames(B)[[1]] = a

###@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
matplot(a, B, type="l", lwd=3, lty=1, main="B Basis")
abline(v=seq(0,60,5),col=grey(.60))


## single year matrix
S = outer(age, a, function(aa,xx) { 1 * (aa <= xx) * (xx < aa+1)  })
dimnames(S) = list(age,a)
S = sweep(S,1,apply(S,1,sum),"/")

#################################
xLim = c( Inf, seq(12.5, 52.5, 5) )
nval = 0:2

Cdata = array(NA, c(2,9,length(xLim), length(nval)) ,
              dimnames=list(c("IDB","HFD"),seq(10,50,5),xLim,nval) )

## IDB is the g=7 case with avg fertility for groups 15-19 ... 45-49
## HFD is the g=9 case with avg fertility for groups 12-14 15-19 ... 50-54

ngroup = c(IDB=7, HFD=9)

for (inData in c("IDB", "HFD")) {  

  if (inData=="HFD") L = seq(10,50,5)
  if (inData=="IDB") L = seq(15,45,5)

  ## function to construct group aggregration matrix G from a vector L
  ## that contains the lower limits of consecutive 5-year age groups
  
  G           = outer(L,a, function(LL,xx) { 1 * (LL <= xx) * (xx < LL+5)  })
  dimnames(G) = list(L,a)
  G           = sweep(G,1,apply(G,1,sum),"/")
  
  VfitInv = 5*Wstar/(.10) * diag(length(L))
  
  C1 =       t(B) %*% t(G) %*% VfitInv %*% G %*% B
  C1 = C1 +  t(B) %*% t(S) %*% t(M)  %*% VshInv %*% M  %*% S %*% B  
    
  C2 = t(B) %*% t(G) %*% VfitInv
  
  Kw = solve(C1) %*% C2
  dimnames(Kw) = list(NULL, L)
  
  Kf    = B %*% Kw

  
  ###@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  matplot(a, Kf, type="l", lwd=3, lty=1, main="Kf Basis")
  abline(v=seq(0,60,5),col=grey(.60))

  ###@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  nn = c(IDB=7, HFD=9)
  matplot(a, Kf[,c(1,ncol(Kf))], type="l", lwd=3, lty=1, 
          main=paste("Kf Basis (g=",nn[inData],") first & last col"),
       ylim=c(-.5,1.5))
  abline(v=seq(0,60,5),h=0,col=grey(.60))
  
  ###@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  
  write.csv( Kw, file=paste("Kw",ngroup[inData],".csv", sep="") )
  write.csv( Kf, file=paste("K",ngroup[inData],".csv", sep="") )

  ### construct the cn[delta] constants for n=0,1,2 and for various conditions
  ###  of interest (a < 15.5, a < 17.5, ..., a < 52.5,  a < Inf)
  
  for (x in xLim) {
    
    for (n in nval) {
        sumvec = 0*L
        for (i in 1:nrow(B)) sumvec = sumvec + (a[i] < x) * a[i]^n * B[i,] %*% Kw * Delta
        Cdata[inData,paste(L),paste(x),paste(n)] = sumvec
    } # for n
    
  } # for xLim
  
  
} # for inData
  

### write the C constants to a file in dput format, and as a big matrix in csv format

## change dimension names to something more general
dimnames(Cdata)[[1]] = c("g=7","g=9")

dput(round(Cdata,3), file="Cdata.dput")

rname = apply( expand.grid(dimnames(Cdata)[c(1,3,4)]), 1, paste, collapse="-")
Cmat = matrix( aperm(round(Cdata,3),c(1,3,4,2)), ncol=9 ,
           dimnames=list(rname,seq(10,50,5)))

write.csv( Cmat, file="Cdata.csv")


